import copy
import json
import os
from threading import Thread
import traceback
from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.gateway.constant_enums import DownlinkMessageType, Status
from pathlib import Path
from zk import ZK, const
from zk.exception import ZKErrorResponse, ZKNetworkError
from datetime import datetime
import time
import pytz
import logging
from simplejson import dumps
import threading
import signal
import sys
from schema import Schema, And, Use, Optional, SchemaError, Or
from functools import reduce

log = logging.getLogger("connector")

sem = threading.Semaphore(1)

MASCK_MAX = 62
MASCK_MIN = 1


def _check_magic_number(magic_number):
    if (magic_number > MASCK_MAX or magic_number < 0):
        raise Exception(
            'Magic number must bigger than {} and smaller than {}'.format(MASCK_MIN, MASCK_MAX))


def _check_user_id_company(user_id_company):
    if (user_id_company >= (MASCK_MIN << 9)):
        raise Exception(
            'Uaser ID from company must be bigger than 1 and smaller than {}'.format(MASCK_MIN))


def is_device_id(magic_number, user_id_device):
    if magic_number == 0x0:
        return True
    return (int(user_id_device) >> 9) == magic_number


def convert_to_device_id(magic_number, user_id_company):
    if magic_number == 0x0:
        return user_id_company
    _check_magic_number(magic_number)
    _check_user_id_company(user_id_company)
    return (magic_number << 9) | user_id_company


def convert_to_company_id(magic_number, user_id_device):
    user_id_device = int(user_id_device)
    if magic_number == 0x0:
        return user_id_device
    _check_magic_number(magic_number)
    return (magic_number << 9) ^ user_id_device


def not_equal_packet(packet_send, packet_save):
    if len(packet_send["telemetry"]) != 0 or packet_save["attributes"] != packet_send["attributes"]:
        return True


ATTENDANCE_TELEMETRY_SCHEMA = Schema({
    Optional('deviceName'): And(str, len),
    Optional('deviceType'): And(str, len),
    Optional('attributes'): [{
        'Device_name': str
    }],
    Optional('telemetry'): [{
        "ts": Or(int, float),
        "values": {
            "user_id": And(int),
            "timestamp":  And(str, len),
            "punch": And(str, len),
            "device_name":  And(str, len),
        }
    }]
}, ignore_extra_keys=True)


class ZktecPro(Connector, Thread):

    def __init__(self, gateway, config, connector_type):
        super().__init__()
        self.statistics = {'MessagesReceived': 0,
                           'MessagesSent': 0}
        self.gateway = gateway  # Reference to TB Gateway
        self.connector_type = connector_type  # Use For Convertor
        self.config = config  # zktec.json Contents
        self.__id = self.config.get('id')

        # Extract main sections from configuration ---------------------------------------------------------------------
        self.__device = config.get('device')
        self.__storage_path = config.get('storage', ".")
        self._name = self.__device.get('name', 'zktec')
        self._ip = self.__device.get('ip', '127.0.0.1')
        self._port = self.__device.get('port', "1883")
        self._password = self.__device.get('password', "1")
        self._timezone = self.__device.get('timezone', "210")
        self._magic_number = int(self.__device.get('magic_number', "0"))

        # Set up lifecycle flags ---------------------------------------------------------------------------------------
        self.connection = False  # Service variable for check connection to device
        self.__logger = logging.getLogger("zkteck-"+self._ip)
        self.__deviceName = config.get('deviceName', 'deviceName')
        self.__deviceType = config.get('deviceType', 'default')
        self.stopped = False 
        self.daemon = True 
        # Create device
        self.gateway.add_device(self.__deviceName, {"connector": self},
                                device_type=self.__deviceType)
        
        self._keep_alive_thread = Thread(target=self._send_keep_alive, daemon=True)
        self._keep_alive_thread.start()

        log.info(f'magic number value is : {self._magic_number}')


    def get_type(self):
        return self.connector_type
    
    def get_id(self):
        return self.__id

    def is_stopped(self):
        return not self.is_connected()
        
    def get_config(self):
        return self.config

    def get_name(self):
        return self.name

    def is_connected(self):
        return self._is_zkteco_connected()

    def run(self):
        while not self.stopped:
            try:
                self._run()
            except Exception as ex:
                logging.error('ZKTec unsupported exception happend: %s', ex)
                traceback.print_exc()
            finally:
                time.sleep(30)

    def _run(self):
        result_dict = {
            'deviceName': self.__deviceName,
            'deviceType': self.__deviceType,
            'attributes': [],
            'telemetry': [],
        }

        # Send Attribute
        result_dict['attributes'].append(self._zkteco_get_attribute())
        device_platform = result_dict['attributes'][0]['Platform']

        # Send Telemetry
        attendances = self._zkteco_get_attendance()
        for attendance in attendances:
            if is_device_id(self._magic_number, attendance.user_id):
                item = self._convert_attendance_to_telemetry(attendance, device_platform)

                if self._should_send_attendance(item):
                    result_dict['telemetry'].append(item)

        # Send result to thingsboard
        if self._must_send_to_storage(result_dict) and self._send_to_storage(result_dict):
            self.PACKET_SAVE = result_dict

    def open(self):
        # TODO: maso, 2023: check the biz follow of open
        self.stopped = False
        self.start()


    def close(self):
    
        log.info(f"Closing ZKTeco connector for device {self.__deviceName}")
        self.stopped = True  
        try:
            self._zkteco_close()
        except Exception as e:
            log.error(f"Error closing ZKTeco connection: {e}")
       
        if self.is_alive():
            self.join(timeout=2.0)  
            if self.is_alive():
                log.warning(f"ZKTeco connector thread {self.__deviceName} did not stop in time")
        log.info(f"ZKTeco connector {self.__deviceName} closed successfully")

    def on_attributes_update(self, content):
        # TODO: maso, 2023: I do not know what is it for?
        pass

    def server_side_rpc_handler(self, content):
        try:
            self._server_side_rpc_handler(content, 3)
        except Exception as ex:
            self.__logger.error("ZKTec unsupported exception happend %s", ex)
            traceback.print_exc()
            self.gateway.send_rpc_reply(
                device=content["device"],
                req_id=content["data"]["id"],
                content={"success_sent": "False",
                         "message": "ZKTec unsupported exception happend %s" % ex}
            )

    def _server_side_rpc_handler(self, content, tries_count=3):
        params = content["data"]["params"]
        method_name = content["data"]["method"]
        try:
            # update_user
            if method_name == "update_user":
                return self.update_user(params, content)

            # delete user
            if method_name == "del_user":
                return self.del_user(params, content)

            # update fingerprint
            if method_name == "update_fingerprint":
                return self.update_fingerprint(params)
        except ZKNetworkError as netex:
            self.__logger.error(
                "ZKTec network error detected : %s, %s", netex, traceback.extract_stack)
            if tries_count > 0:

                return self._server_side_rpc_handler(content, tries_count-1)
            raise netex

    def update_fingerprint(self, params):
        magic_user_id = convert_to_device_id(
            user_id_company=int(params["user_id_change"]),
            magic_number=self._magic_number
        )
        self._zkteco_enroll_user(uid=magic_user_id)

    def update_user(self, params, content):
        users = self._zkteco_get_users()

        for key, value in params.items():
            # Check user is exist
            exist_user = 0
            # set magic number
            magic_user_id = convert_to_device_id(
                user_id_company=int(value["uid"]),
                magic_number=self._magic_number)

            for item in users:
                if item.uid == magic_user_id:
                    exist_user = item

            # user not exist Create user
            if exist_user == 0:
                self._zkteco_set_user(uid=int(magic_user_id),
                                      name=value["name"],
                                      privilege=value["privilege"],
                                      password=value["password"],
                                      group_id=value["group_id"],
                                      card=int(value["card"]))
            else:
                # user is exist delete and create
                # save finger print

                fingers = self._zkteco_get_templates()

                save_fingers = []
                for finger in fingers:
                    if finger.uid == magic_user_id:
                        save_fingers.append(finger)

                self._zkteco_delete_user(user_id=magic_user_id)
                self._zkteco_set_user(uid=int(magic_user_id),
                                      name=value["name"],
                                      privilege=value["privilege"],
                                      password=value["password"],
                                      group_id=value["group_id"],
                                      card=int(value["card"]))
                # add finger print
                self._zkteco_save_user_template(magic_user_id, save_fingers)

            self._gateway_send_rpc_reply(
                device=content["device"],
                req_id=content["data"]["id"],
                content={"success_sent": 'True'}
            )

    def del_user(self, params, content):

        for key, value in params.items():
            magic_user_id = convert_to_device_id(
                user_id_company=int(value['user_id_delete']),
                magic_number=self._magic_number)

            self._zkteco_delete_user(magic_user_id)

            self.gateway.send_rpc_reply(
                device=content["device"],
                req_id=content["data"]["id"],
                content={"success_sent": 'True'}
            )

    ##############################################################################################
    #                                 packets utils
    ##############################################################################################
    PACKET_SAVE = {'attributes': [],
                   'telemetry': []}

    def _must_send_to_storage(self, result_dict):
        # Check Successful Send
        if not not_equal_packet(result_dict, self.PACKET_SAVE):
            return False

        try:
            ATTENDANCE_TELEMETRY_SCHEMA.validate(result_dict)
        except SchemaError as ex:
            self.__logger.error("ZKTec unsupported exception happend %s", ex)
            traceback.print_exc()
            return False

        return True

    def _should_send_attendance(self, item):
        lastdatetime = self.lastdatetime_text_file()
        # TODO: maso, 2023: check last time stamp

        return item['ts'] > lastdatetime

    def _convert_attendance_to_telemetry(self, attendance, device_platform):
        tz = pytz.FixedOffset(int(self._timezone))

        datetimeOrg = attendance.timestamp
        dateTimeUts = datetime(
            datetimeOrg.year,
            datetimeOrg.month,
            datetimeOrg.day,
            datetimeOrg.hour,
            datetimeOrg.minute,
            datetimeOrg.second,
            tzinfo=tz
        )
        attendance_date = datetime.fromtimestamp(
            datetime.timestamp(dateTimeUts))

        _punch = "in"
        if device_platform == "ZMM220_TFT" and attendance.punch == 2:
            _punch = "out"
        elif device_platform == "ZEM600_TFT" and attendance.punch == 1:
            _punch = "out"

        attendance_telemetry = {
            "ts": attendance_date.timestamp()*1000,
            "values": {
                "user_id": convert_to_company_id(
                    magic_number=self._magic_number,
                    user_id_device=int(attendance.user_id)
                ),
                "timestamp": str(attendance_date),
                "punch": _punch,
                "device_name": self.__deviceName
            }
        }

        return attendance_telemetry
    # Main method of thread, must contain an infinite loop and all calls to data receiving/processing functions.

    def get_storage_path(self):
        path_storage = Path(self.__storage_path + '/%s.txt' %
                            self.config.get('deviceName'))
        return path_storage

    ##############################################################################################
    #                             zkteck connection utils
    ##############################################################################################

    def _is_zkteco_connected(self):
        a = self._zkteco_get_attribute()
        if a:
            return True
        else:
            return False


    def _zkteco_close(self):
      
        try:
            if self.connection is not None:
                self.connection.disconnect()
                log.debug(f"ZKTeco connection for {self.__deviceName} disconnected")
        except Exception as e:
            log.error(f"Error disconnecting ZKTeco device {self.__deviceName}: {e}")
        finally:
            self.connection = None

    def _zkteco_connect(self):
        """connects to the device throu tcp

        raise excpetion if connection fail
        """
        # if self.connection and not self.connection.disconnect():
        #     raise Exception("Fail to disconnect from device. Something bad happend.")

        self.connection = ZK(
            self._ip,
            port=int(self._port),
            timeout=5,
            password=int(self._password),
            force_udp=False,
            ommit_ping=True,
            verbose=False,
            encoding="cp1256"

        )
        if not self.connection.connect():
            raise Exception("Fail to connect to the device")

    ##############################################################################################
    #                                 zkteck util
    ##############################################################################################

    def _zkteco_get_attribute(self):
        try:
            sem.acquire()
            self._zkteco_connect()
            try:
                self.connection.read_sizes()
            except:
                logging.info("Usage Space device Not available")
            try:
                firmware_version = self.connection.get_firmware_version()
            except:
                logging.info("Firmware Version device Not available")
            try:
                serialnumber = self.connection.get_serialnumber()
            except:
                logging.info("Serial Number device Not available")
            try:
                platform = self.connection.get_platform()
            except:
                logging.info("Platform device Not available")
            try:
                device_name = self.connection.get_device_name()
            except:
                logging.info("Device Name Not available")
            try:
                face_version = self.connection.get_face_version()
            except:
                logging.info("Face Version device Not available")
            try:
                fp_version = self.connection.get_fp_version()
            except:
                logging.info("Finter Print Version device Not available")
            try:
                extend_fmt = self.connection.get_extend_fmt()
            except:
                logging.info("Extend FMT device Not available")
            try:
                user_extend_fmt = self.connection.get_user_extend_fmt()
            except:
                logging.info("User Extend FMT device Not available")
            try:
                face_fun_on = self.connection.get_face_fun_on()
            except:
                logging.info("Face Fun On device Not available")
            try:
                compat_old_firmware = self.connection.get_compat_old_firmware(),
            except:
                logging.info("Compat Old Firmware device Not available")
            try:
                network_params = self.connection.get_network_params()
            except:
                logging.info("Network Params device Not available")
            try:
                mac = self.connection.get_mac()
            except:
                logging.info("Mac device Not available")
            try:
                pin_width = self.connection.get_pin_width()
            except:
                logging.info("pin_width device Not available")

            return {
                "Records": self.connection.records,
                "Max_Records": self.connection.rec_cap,
                "Users": self.connection.users,
                "Max_Users": self.connection.users_cap,
                "Fingers": self.connection.fingers,
                "Max_Fingers": self.connection.fingers_cap,
                "Faces": self.connection.faces,
                "Max_Faces": self.connection.faces_cap,
                "Firmware_Version": firmware_version,
                "Serialnumber": serialnumber,
                "Platform": platform,
                "Device_name": device_name,
                "Face_Version": face_version,
                "Finger_Print_Version": fp_version,
                "Extend_Fmt": extend_fmt,
                "User_Extend_Fmt": user_extend_fmt,
                "Face_Fun_On": face_fun_on,
                "Compat_Old_Firmware": compat_old_firmware,
                "Network_Params": network_params,
                "Mac": mac,
                "Pin_Width": pin_width
            }
        finally:
            self._zkteco_close()
            sem.release()

    def _zkteco_get_users(self):
        sem.acquire()
        self._zkteco_connect()
        try:
            return self.connection.get_users()
        finally:
            self._zkteco_close()
            sem.release()

    def _zkteco_delete_user(self, user_id):
        sem.acquire()
        self._zkteco_connect()
        try:
            self.connection.delete_user(uid=user_id)
        finally:
            self._zkteco_close()
            sem.release()

    def _zkteco_set_user(self, *args, **kwargs):
        sem.acquire()
        self._zkteco_connect()
        try:
            self.connection.set_user(*args, **kwargs)
        finally:
            self._zkteco_close()
            sem.release()

    def _zkteco_get_attendance(self):
        sem.acquire()
        self._zkteco_connect()
        try:
            return self.connection.get_attendance()
        finally:
            self._zkteco_close()
            sem.release()

    def _zkteco_get_templates(self):
        sem.acquire()
        self._zkteco_connect()
        try:
            return self.connection.get_templates()
        finally:
            self._zkteco_close()
            sem.release()

    def _zkteco_save_user_template(self, *args, **kwargs):
        sem.acquire()
        self._zkteco_connect()
        try:
            self.connection.save_user_template(*args, **kwargs)
        finally:
            self._zkteco_close()
            sem.release()

    def _zkteco_enroll_user(self, *args, **kwargs):

        sem.acquire()
        self._zkteco_connect()
        try:
            self.connection.enroll_user(*args, **kwargs)
        finally:
            self._zkteco_close()
            sem.release()

    def _send_keep_alive(self):
        KEEP_ALIVE_INTERVAL = 300
        while not self.stopped:
            try:
                keep_alive_data = {
                    "deviceName": self.__deviceName,
                    "deviceType": self.__deviceType,
                    "attributes": [
                        {"status": "active"}
                    ],
                    "telemetry": [
                        {
                            "ts": int(time.time() * 1000), 
                            "values": {
                                "keep_alive": True,
                                "device_name": self.__deviceName
                            }
                        }
                    ]
                }
                if self._send_to_storage(keep_alive_data):
                    self.__logger.info(f"Sent keep-alive message for device {self.__deviceName}")
                else:
                    self.__logger.warning(f"Failed to send keep-alive message for device {self.__deviceName}")
            except Exception as e:
                self.__logger.error(f"Error sending keep-alive message: {e}")
            for _ in range(int(KEEP_ALIVE_INTERVAL / 10)):
                if self.stopped:
                    break
                time.sleep(10)       
            

    ##############################################################################################
    #                                 Gateway util
    ##############################################################################################

    def _gateway_send_rpc_reply(self, *args, **kwargs):
        self.gateway.send_rpc_reply(*args, **kwargs)

    def _send_to_storage(self, result_dict):
        if self.gateway.send_to_storage(self.get_name(), self.get_id(), result_dict) == Status.SUCCESS:
            init_value = {
                'ts': 0
            }
            latest = reduce(
                lambda x, y: x if x['ts'] > y['ts'] else y, result_dict['telemetry'], init_value)
            if latest['ts'] == 0:
                return True
            path = self.get_storage_path()
            if not os.path.exists(os.path.dirname(path.absolute())):
                try:
                    os.makedirs(os.path.dirname(path.absolute()))
                except OSError as exc:  # Guard against race condition
                    raise
            with open(path, 'w') as f:
                f.write(str(latest['ts']))
            return True
        return False

    def lastdatetime_text_file(self):
        # Fetch last date time
        try:
            with open(self.get_storage_path(), 'r') as f:
                lines = f.readlines()
                return float(lines[0])
        except:
            return 0



import threading
import traceback
import cups  
from threading import Thread
from thingsboard_gateway.connectors.connector import Connector, log
import time
import os
import logging
from collections import Counter
import sys
import base64
import tempfile

sem = threading.Semaphore(1)

PACKET_SAVE ={'attributes': [],
                'telemetry': []}


def equal_packet(packet_send, packet_save):
    if packet_save["telemetry"] == packet_send["telemetry"] and packet_save["attributes"] == packet_send["attributes"]:
        return True
    else:
        return False
      
class CUPSPro(Connector, Thread): 
    # TODO : IPPError, HTTPEttor, check try except
    
    def __init__(self, gateway, config, connector_type):
        super().__init__()
        self.statistics = {'MessagesReceived': 0,
                           'MessagesSent': 0}
        self.gateway = gateway  # Reference to TB Gateway
        self.connector_type = connector_type  # Use For Convertor
        self.config = config  # sii.json Contents

        # Extract main sections from configuration ---------------------------------------------------------------------
        self.__prefix = config.get('prefix', 'prefix')
        self.__deviceName = config.get('deviceName', 'deviceName')
        self.__deviceType = config.get('deviceType', 'default')
        
        # XXX: maso, 2023: rais error if prefix is not set
        # Set up lifecycle flags ---------------------------------------------------------------------------------------
        self.connection = False  # Service variable for check connection to device
        self.__logger = logging.getLogger("cups-"+self.__deviceName)
        
        # Create device
        self.connection = cups.Connection()
        self.printers = self.connection.getPrinters()
                
        for key, config in self.printers.items():
            self.gateway.add_device(self.__prefix + key, {"connector": self},
                                    device_type=self.__deviceType)
 
    def connect_device(self):
        """connects to the device throu tcp
        
        raise excpetion if connection fail
        """
        try:
            self.connection = cups.Connection()
            self.printers = self.connection.getPrinters()
                
        except (cups.IPPError) as e:
            pass
           
    def open(self): 
        self.stopped = False
        self.start()

    def get_name(self):
        return self.name

    def is_connected(self):
        return self.connection 
    
    def find_jobs(self):
        # Find jobs
        dict_jobs_counter = {}
        job_completed = self.connection.getJobs(which_jobs='completed')
        job_not_completed = self.connection.getJobs(which_jobs='not-completed')
        job_all = self.connection.getJobs(which_jobs='all')

        # Completed Jobs
        result_job_completed= []
        for job in job_completed:
            string = self.connection.getJobAttributes(job)['job-printer-uri']
            printer_name_for_select_job = list(string.split("/"))[-1]
            result_job_completed.append(printer_name_for_select_job)
        dict_count_job_completed = Counter(result_job_completed)

        # Not_Completed Jobs
        result_job_not_completed= []
        for job in job_not_completed:
            string = self.connection.getJobAttributes(job)['job-printer-uri']
            printer_name_for_select_job = list(string.split("/"))[-1]
            result_job_not_completed.append(printer_name_for_select_job)
        dict_count_job_not_completed = Counter(result_job_not_completed)

        # All Jobs
        result_job_all= []
        for job in job_all:
            string = self.connection.getJobAttributes(job)['job-printer-uri']
            printer_name_for_select_job = list(string.split("/"))[-1]
            result_job_all.append(printer_name_for_select_job)
        dict_count_job_all = Counter(result_job_all)

        # Add Counter jobs
        dict_jobs_counter["Completed_jobs"] = dict_count_job_completed
        dict_jobs_counter["not_completed_jobs"] = dict_count_job_not_completed
        dict_jobs_counter["all_jobs"] = dict_count_job_all
        
        return dict_jobs_counter
     
    # Main method of thread, must contain an infinite loop and all calls to data receiving/processing functions.
    def run(self):   
        
        while (True):
            try:
                if not self.is_connected():
                    self.connect_device()
                    
                # Find jobs
                dict_count_job_all = self.find_jobs() 
                
                for key, config in self.printers.items():
                    self.result_dict = {
                        'deviceName': self.__prefix + key,
                        'deviceType': self.__deviceType,
                        'attributes': [],
                        'telemetry': [],
                    }
                    
                    # Send attribute 
                    for attribute_key in ['printer-is-shared', 'printer-state', 'printer-state-message', 'printer-state-reasons', 'printer-type',
                                          'printer-uri-supported', 'printer-location', 'printer-info', 'device-uri', 'printer-make-and-model']:
                        self.result_dict['attributes'].append({
                            attribute_key: config[attribute_key]
                        })
                        
                    # Send telemetry
                    for key_counter , value_counter in dict_count_job_all.items():
                        self.result_dict['telemetry'].append({
                        key_counter: value_counter[key]
                    })

                    if not equal_packet(self.result_dict,PACKET_SAVE):
                        # Send result to thingsboard
                        self.gateway.send_to_storage(self.get_name(), self.result_dict)
                        PACKET_SAVE["attributes"] = self.result_dict["attributes"]
                        PACKET_SAVE["telemetry"] = self.result_dict["telemetry"]
                        
            except:
                logging.error("Printer Not Connected")
                
            
            time.sleep(30)
    
    def close(self):
        #printer_names = self.connection.printers.keys()
        #for printer_name in printer_names:    
        #        self.connection.disablePrinter(printer_name)
        #signal.signal(signal.SIGINT, self.signal_handler)
        #signal.signal(signal.SIGINT, self.stop())
        pass
        
    def on_attributes_update(self, content):
        pass
    
    def print_file(self, printer_name , printer_file): 
        printid = self.connection.printFile(printer_name, printer_file ,'' , {})
        
    def server_side_rpc_handler(self, content):
        sem.acquire()
        try:
            self._server_side_rpc_handler(content, 3)
        except Exception as ex :
            self.__logger.error("Cups unsupported exception happend %s", ex)
            traceback.print_exc()
            
            self.gateway.send_rpc_reply(
                device= content["device"], 
                req_id= content["data"]["id"],
                content = {"success_sent":"False" , "message" :"Cups unsupported exception happend %s" % ex }
            )
        finally: 
            sem.release()
            time.sleep(0.25)

    def _server_side_rpc_handler(self, content, tries_count=3):
        params = content["data"]["params"]
        method_name = content["data"]["method"]
        
        # printer_file is encode string so first decode string
        try:
            if method_name == "print_file":
                # create temporary directory
                temp = tempfile.NamedTemporaryFile(suffix=params["suffix_file_name"])
                
                try:
                    coded_string = params["content"]
                    base64_decode = base64.b64decode(coded_string)
                    text_file = open(temp.name, "wb")
                    n = text_file.write(base64_decode)
                    self.print_file(content["device"].replace(self.__prefix, ""), temp.name)
                    #content["device"].replace(self.__prefix, "")
                finally:
                    temp.close()
    
                
        except Exception as ex :
            self.__logger.error("Cups unsupported exception happend : %s, %s", ex, traceback.extract_stack)
            if tries_count > 0:
                self.connect_device()
                return self._server_side_rpc_handler(content, tries_count-1)
            raise ex
            



import sys
import signal
import logging

import rich
from zk import ZK, const


def zk_devices_check_connection(ip=None, password=None, port=4370):
    """Connects to the ZK device and get more information from device.
    """
    logging.debug("create ZK instance")
    zk = ZK(str(ip), port=port, timeout=5, password=int(
        password), force_udp=False, ommit_ping=True)

    try:
        logging.debug("Connect to device")
        conn = zk.connect()
        attendances = conn.get_attendance()
        users = conn.get_users()
        rich.print(users)
        rich.print(attendances)
    finally:
        if conn:
            conn.disconnect()


def zk_devices_check_utf8(ip=None, password=None, port=4370):
    """Check devices language
    """
    conn = None
    logging.debug("create ZK instance")
    zk = ZK(str(ip), port=port, timeout=5, password=int(
        password), force_udp=False, ommit_ping=True)

    try:
        logging.debug("Connect to device")
        conn = zk.connect()
        # disable device, this method ensures no activity on the device while the
        # process is run
        conn.disable_device()
        # another commands will be here!
        conn.set_user(uid=1000, name='فارسی', privilege=const.USER_DEFAULT,
                      password='', group_id='', user_id='1000', card=0)

    except Exception:
        logging.error("Fail to get attendences", exc_info=True)
    finally:
        if conn:
            conn.disconnect()


def zk_devices_get_attendance(ip=None, password=None, port=4370):
    """Check devices language
    """
    logging.debug("create ZK instance")
    zk = ZK(str(ip), port=port, timeout=5, password=int(
        password), force_udp=False, ommit_ping=True)

    try:
        logging.debug("Connect to device")
        conn = zk.connect()
        # disable device, this method ensures no activity on the device while the process is run
        conn.disable_device()
        # another commands will be here!
        # Example: Get All Users
        attendances = conn.get_attendance()
        rich.print(attendances)
    finally:
        if conn:
            conn.disconnect()


def zk_devices_get_user_info(ip=None, password=None, port=4370):
    """Check devices language
    """
    logging.debug("create ZK instance")
    zk = ZK(str(ip), port=port, timeout=5, password=int(
        password), force_udp=False, ommit_ping=True)

    try:
        logging.debug("Connect to device")
        conn = zk.connect()
        # disable device, this method ensures no activity on the device while the process is run
        conn.disable_device()
        # another commands will be here!
        # Example: Get All Users
        users = conn.get_users()
        for user in users:
            privilege = 'User'
            if user.privilege == const.USER_ADMIN:
                privilege = 'Admin'
            rich.print('+ UID #{}'.format(user.uid))
            rich.print('  Name       : {}'.format(user.name))
            rich.print('  Privilege  : {}'.format(privilege))
            rich.print('  Password   : {}'.format(user.password))
            rich.print('  Group ID   : {}'.format(user.group_id))
            rich.print('  User  ID   : {}'.format(user.user_id))

        # Test Voice: Say Thank You
        conn.test_voice()
        # re-enable device after all commands already executed
        conn.enable_device()
    finally:
        if conn:
            conn.disconnect()



import zktec_connector
import unittest
from unittest.mock import MagicMock
import sys
import os
from thingsboard_gateway.gateway.constant_enums import DownlinkMessageType, Status
from datetime import datetime

from zk import attendance

sys.path.append(os.path.abspath(
    '/home/sanaz/viraweb123/odoo-iot/vw-gateway/extensions/attendence/zktec'))


class Object(object):
    pass


class TestStringMethods(unittest.TestCase):

    def test_convert_to_device_id(self):
        data = [
            (1, 31745, 62),
            (1, 0x3C01, 30),
            (1, 0x201,  1),
            (1, 513,  1),
            # Magic number is default
            (513, 513, 0),
            (4095, 4095, 0x0),
        ]
        for user_id, device_id, magic_number in data:
            self.assertEqual(zktec_connector.convert_to_device_id(magic_number, user_id),
                             device_id)
            self.assertEqual(zktec_connector.convert_to_company_id(magic_number, device_id),
                             user_id)

    def test_convert_to_device_id_invalid(self):
        ids = [
            2500,
            6000,
            2049,
        ]
        magic_number = 0x7c00
        for id in ids:
            self.assertRaises(
                Exception, zktec_connector.convert_to_device_id, magic_number, id)

    def test_convert_to_device_id_invalid_magic_number(self):
        id = 0x01
        magic_number = 0x8c00
        self.assertRaises(
            Exception, zktec_connector.convert_to_device_id, magic_number, id)

    def test_send_to_storage(self):
        config = Object()
        device = Object()
        gateway = Object()
        connector_type = Object()

        device.get = MagicMock(return_value='1')
        gateway.send_to_storage = MagicMock(return_value=Status.SUCCESS)
        gateway.add_device = MagicMock(return_value=True)

        def side_effect_func(val, *args, **kwargs):
            if val == 'device':
                return device
            return 'test'
        config.get = MagicMock(side_effect=side_effect_func)

        connction = zktec_connector.ZktecPro(gateway, config, connector_type)
        self.assertIsNotNone(connction)

        result_dict = {
            'telemetry': [{
                "ts": 1,
                "values": {
                    "user_id": 3,
                    "timestamp":  "2023-10-4 10:01:00",
                    "punch": "out",
                    "device_name":  "test2"}
            }, {
                "ts": 2,
                "values": {
                    "user_id": 3,
                    "timestamp":  "2023-10-4 10:02:00",
                    "punch": "out",
                    "device_name":  "test2"}
            }, {
                "ts": 3,
                "values": {
                    "user_id": 3,
                    "timestamp":  "2023-10-4 10:03:00",
                    "punch": "out",
                    "device_name":  "test2"}
            }]
        }
        path = connction.get_storage_path()
        path.unlink(missing_ok=True)

        connction._send_to_storage(result_dict)
        self.assertTrue(path.is_file())

        self.assertEqual(3, connction.lastdatetime_text_file())
        path.unlink(missing_ok=True)

    def test_lastdatetime_text_file(self):
        config = Object()
        device = Object()
        gateway = Object()
        connector_type = Object()

        device.get = MagicMock(return_value='1')
        gateway.send_to_storage = MagicMock(return_value=Status.SUCCESS)
        gateway.add_device = MagicMock(return_value=True)

        def side_effect_func(val, *args, **kwargs):
            if val == 'device':
                return device
            return 'test'
        config.get = MagicMock(side_effect=side_effect_func)

        connction = zktec_connector.ZktecPro(gateway, config, connector_type)
        self.assertIsNotNone(connction)

        path = connction.get_storage_path()
        path.unlink(missing_ok=True)
        self.assertEqual(0, connction.lastdatetime_text_file())
        path.unlink(missing_ok=True)

    def test_run_single(self):
        config = Object()
        device = Object()
        gateway = Object()
        connector_type = Object()

        device.get = MagicMock(return_value='1')
        gateway.send_to_storage = MagicMock(return_value=Status.SUCCESS)
        gateway.add_device = MagicMock(return_value=True)

        def side_effect_func(val, *args, **kwargs):
            if val == 'device':
                return device
            return 'test'
        config.get = MagicMock(side_effect=side_effect_func)

        connction = zktec_connector.ZktecPro(gateway, config, connector_type)
        self.assertIsNotNone(connction)

        connction._zkteco_connect = MagicMock(return_value=True)
        connction._zkteco_get_attribute = MagicMock(return_value={
            "ZKTec Error": False,
            "Records": 100,
            "Max Records": 1000,
            "Users": 10,
            "Max Users": 1000,
            "Fingers": 20,
            "max_fingers": 2000,
            "Faces": 10,
            "Max Faces": 1000,
            "Firmware Version": "version",
            "Serialnumber": "1234558",
            "Platform": "platform",
            "Device_name": "device_name",
            "Face Version": "face_version",
            "Finger Print Version": "fp_version",
            "Extend Fmt": "extend_fmt",
            "User Extend Fmt": "user_extend_fmt",
            "Face Fun On": "face_fun_on",
            "Compat Old Firmware": "compat_old_firmware",
            "Network Params": "network_params",
            "Mac": "mac",
            "Pin Width": "pin_width"
        })
        connction._zkteco_get_attendance = MagicMock(return_value=[])

        connction._run()
        self.assertEqual(gateway.send_to_storage.call_count, 1)

        connction._run()
        self.assertEqual(gateway.send_to_storage.call_count, 1)

    def test_run_single_with_telemetry(self):
        config = Object()
        device = Object()
        gateway = Object()
        connector_type = Object()

        device.get = MagicMock(return_value='0')
        gateway.send_to_storage = MagicMock(return_value=Status.SUCCESS)
        gateway.add_device = MagicMock(return_value=Status.SUCCESS)

        def side_effect_func(val, *args, **kwargs):
            if val == 'device':
                return device
            if val == 'magic_number':
                return 3
            return 'test'
        config.get = MagicMock(side_effect=side_effect_func)

        connction = zktec_connector.ZktecPro(gateway, config, connector_type)
        self.assertIsNotNone(connction)

        connction._zkteco_connect = MagicMock(return_value=True)
        connction._zkteco_get_attribute = MagicMock(return_value={
            "ZKTec Error": False,
            "Records": 100,
            "Max Records": 1000,
            "Users": 10,
            "Max Users": 1000,
            "Fingers": 20,
            "max_fingers": 2000,
            "Faces": 10,
            "Max Faces": 1000,
            "Firmware Version": "version",
            "Serialnumber": "1234558",
            "Platform": "platform",
            "Device_name": "device_name",
            "Face Version": "face_version",
            "Finger Print Version": "fp_version",
            "Extend Fmt": "extend_fmt",
            "User Extend Fmt": "user_extend_fmt",
            "Face Fun On": "face_fun_on",
            "Compat Old Firmware": "compat_old_firmware",
            "Network Params": "network_params",
            "Mac": "mac",
            "Pin Width": "pin_width"
        })
        connction._zkteco_get_attendance = MagicMock(return_value=[
            attendance.Attendance(user_id=1, timestamp=datetime(
                2022, 1, 1, 8, 0, 1), status=1, punch=0, uid=1),
            attendance.Attendance(user_id=1, timestamp=datetime(
                2022, 1, 1, 9, 0, 1), status=1, punch=1, uid=1),
            attendance.Attendance(user_id=1, timestamp=datetime(
                2022, 1, 1, 10, 0, 1), status=1, punch=0, uid=1),
            attendance.Attendance(user_id=1, timestamp=datetime(
                2022, 1, 1, 11, 0, 1), status=1, punch=1, uid=1),
        ])

        connction._run()
        self.assertEqual(gateway.send_to_storage.call_count, 1)

        connction._run()
        self.assertEqual(gateway.send_to_storage.call_count, 1)

    def test_run_telemetry_with_magic_number(self):
        config = Object()
        device = Object()
        gateway = Object()
        connector_type = Object()

        def side_effect_func_device(val, *args, **kwargs):
            if val == 'magic_number':
                return '1'
            return '0'
        device.get = MagicMock(side_effect=side_effect_func_device)
        gateway.send_to_storage = MagicMock(return_value=Status.SUCCESS)
        gateway.add_device = MagicMock(return_value=Status.SUCCESS)

        def side_effect_func(val, *args, **kwargs):
            if val == 'device':
                return device
            if val == 'magic_number':
                return 1
            return 'test'
        config.get = MagicMock(side_effect=side_effect_func)

        connector = zktec_connector.ZktecPro(gateway, config, connector_type)
        self.assertIsNotNone(connector)

        attendances = [
            attendance.Attendance(user_id=530, timestamp=datetime(
                2022, 1, 1, 8, 0, 1), status=1, punch=0,  uid=530),
            attendance.Attendance(user_id=530, timestamp=datetime(
                2022, 1, 1, 9, 0, 1), status=1, punch=1,  uid=530),
            attendance.Attendance(user_id=530, timestamp=datetime(
                2022, 1, 1, 10, 0, 1), status=1, punch=0, uid=530),
            attendance.Attendance(user_id=530, timestamp=datetime(
                2022, 1, 1, 11, 0, 1), status=1, punch=1, uid=530),
        ]
        conveted_attendance = connector._convert_attendance_to_telemetry(
            attendances[0], "ZMM220_TFT")
        self.assertEqual(conveted_attendance['values']['user_id'], 18)

if __name__ == '__main__':
    unittest.main()





import typer
from typing_extensions import Annotated

from tb_gateway.zkteco import devices

app = typer.Typer()


@app.command()
def check(
    ip: Annotated[str, typer.Option(help="IP address of the device.")],
    password: Annotated[str, typer.Option(help="Devise access key.")],
):
    """Checks connection to the ZK devices
    """
    devices.zk_devices_check_connection(ip, password)


@app.command("check-utf")
def check_utf(
    ip: Annotated[str, typer.Option(help="IP address of the device.")],
    password: Annotated[str, typer.Option(help="Devise access key.")],
):
    """Check if possible to write UTF8 text
    """
    devices.zk_devices_check_utf8(ip, password)


@app.command("get-attendance")
def _get_attendance(
    ip: Annotated[str, typer.Option(help="IP address of the device.")],
    password: Annotated[str, typer.Option(help="Devise access key.")],
):
    """Get attendances"""
    devices.zk_devices_get_attendance(ip, password)


@app.command("user-info")
def _show_user_info_cmd(
    ip: Annotated[str, typer.Option(help="IP address of the device.")],
    password: Annotated[str, typer.Option(help="Devise access key.")],
):
    """Shows user infor"""
    devices.zk_devices_get_user_info(ip, password)


if __name__ == "__main__":
    app()



import unittest
import os


if __name__ == '__main__':
    loader = unittest.TestLoader()
    start_dir = os.path.dirname(os.path.abspath(__file__))
    suite = loader.discover(start_dir)

    runner = unittest.TextTestRunner()
    runner.run(suite)


import unittest
import sys
import os
from schema import Schema, And, Use, Optional, SchemaError 

sys.path.append(os.path.abspath('/home/sanaz/viraweb123/odoo-iot/vw-gateway/extensions/attendence/zktec'))
import zktec_connector

ATTENDANCE_TELEMETRY_SCHEMA = Schema({
    "ts": And(int),
    "values": {
        "user_id": And(int),
        "timestamp":  And(str, len),
        "punch": And(str, len),
        "device_name":  And(str, len),
    }
})


class TestStringMethods(unittest.TestCase):

    def test_schema(self):
        data = {
            'telemetry': [{
                "ts": 1,
                "values": {
                    "user_id": 2,
                    "timestamp":  "2023-10-3 10:20:05",
                    "punch": "in",
                    "device_name":  "test", }},
                {
                "ts": 2,
                "values": {
                    "user_id": 3,
                    "timestamp":  "2023-10-4 10:20:05",
                    "punch": "out",
                    "device_name":  "test2"}}
            ]
        }
        valid_date = zktec_connector.ATTENDANCE_TELEMETRY_SCHEMA.validate(data)
        self.assertIsNotNone(valid_date)

    def test_schema_invalid(self):
        data = {
            'telemetry': [{
                "ts": 1,
                "values": {
                    "user_id": "2",
                    "timestamp":  "2023-10-3 10:20:05",
                    "punch": "in",
                    "device_name":  "test",}},
                {
                "ts": 2,
                "values": {
                    "timestamp":  "2023-10-4 10:20:05",
                    "punch": "out",
                    "device_name":  "test2"}}
                ]}
        with self.assertRaises(SchemaError):
            zktec_connector.ATTENDANCE_TELEMETRY_SCHEMA.validate(data)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
from thingsboard_gateway.gateway.tb_gateway_service import TBGatewayService
from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader
import os.path
from pyfiglet import Figlet
import pkg_resources
import typer
import signal
import sys
import logging
import shutil
from pathlib import Path
import importlib.resources as resources
import tb_gateway.vw_gateway.templates as templates

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

app = typer.Typer()


gateway = None


def signal_handler(sig, frame):
    log.info("Received SIGINT, stopping the gateway...")
    global gateway
    if gateway is not None:
        try:
            gateway.stop()
            log.info("Gateway stopped successfully.")
        except Exception as e:
            log.error(f"Error stopping gateway: {e}")
    else:
        log.warning("Gateway instance not initialized.")
    sys.exit(0)


# signal.signal(signal.SIGINT, signal_handler)


@app.command()
def main(path_cofig: str, path_extension: str):
    f = Figlet(font='big')
    print(f.renderText('Odoonix'))
    print("Thingsboard Gateway Package Version is: " +
          pkg_resources.get_distribution("thingsboard-gateway").version)
    TBModuleLoader.PATHS.append(path_extension)
    global gateway
    gateway = TBGatewayService(path_cofig)

    signal.pause()


TEMPLATE_FILES = [
    "conf.json",
    "logs.json",
    "statistics.json",
    "devices/cups.json",
    "devices/zktec.json",
]

@app.command()
def init():
    output_dir = Path.cwd()
    files_root = resources.files(templates)

    for relative_path_str in TEMPLATE_FILES:
        relative_path = Path(relative_path_str)
        file_in_package = files_root / relative_path

        if not file_in_package.is_file():
            typer.echo(f"Skipping: {relative_path} (not a file)")
            continue

        with resources.as_file(file_in_package) as real_file:
            destination = output_dir / relative_path
            destination.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(real_file, destination)
            typer.echo(f"Copied: {relative_path}")

    typer.echo("Selected template files copied successfully.")


if __name__ == "__main__":
    app()
